/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.query.id;

/**
 * Generator used to provide query ids.
 */
public interface QueryIdGenerator {

  /**
   * Returns the next query id identifier.
   *
   * @return a string query id identifier
   */
  String getNext();

  /**
   * Create an generator that can be used to create new ids without affecting the state
   * of the real generator.
   *
   * @return a new generator initialized at what the next id would be
   */
  QueryIdGenerator createSandbox();
}
