/*
 * Copyright 2021 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.execution.interpreter.terms;

import com.google.common.collect.ImmutableList;
import io.confluent.ksql.execution.codegen.helpers.ArrayBuilder;
import io.confluent.ksql.execution.interpreter.TermEvaluationContext;
import io.confluent.ksql.schema.ksql.types.SqlType;
import java.util.List;

public class CreateArrayTerm implements Term {

  private final List<Term> arrayTerms;
  private final SqlType resultType;

  public CreateArrayTerm(final List<Term> arrayTerms, final SqlType resultType) {
    this.arrayTerms = ImmutableList.copyOf(arrayTerms);
    this.resultType = resultType;
  }

  @Override
  public Object getValue(final TermEvaluationContext context) {
    final ArrayBuilder build = new ArrayBuilder(arrayTerms.size());
    for (Term term : arrayTerms) {
      build.add(term.getValue(context));
    }
    return build.build();
  }

  @Override
  public SqlType getSqlType() {
    return resultType;
  }
}
